/**
 * <copyright>
 * 
 * Copyright (c) 2013 itemis and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *     itemis - Initial API and implementation
 * 
 * </copyright>
 */
package org.safe.ide.scoping;

import java.util.Collection;
import java.util.Map;
import java.util.WeakHashMap;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.common.util.BasicDiagnostic;
import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.osgi.util.NLS;
import org.eclipse.sphinx.emf.metamodel.IMetaModelDescriptor;
import org.eclipse.sphinx.emf.metamodel.MetaModelDescriptorRegistry;
import org.eclipse.sphinx.emf.scoping.IResourceScope;
import org.eclipse.sphinx.emf.scoping.ProjectResourceScope;
import org.eclipse.sphinx.emf.scoping.ProjectResourceScopeProvider;
import org.eclipse.sphinx.emf.scoping.ResourceScopeMarkerSynchronizer;
import org.eclipse.sphinx.platform.resources.DefaultResourceChangeHandler;
import org.eclipse.sphinx.platform.resources.ResourceDeltaVisitor;
import org.eclipse.sphinx.platform.util.PlatformLogUtil;
import org.safe.ide.metamodel.SAFEReleaseDescriptor;
import org.safe.ide.internal.Activator;
import org.safe.ide.internal.messages.Messages;
import org.safe.ide.preferences.ISAFEWorkspacePreferences;

public class SAFEProjectResourceScopeProvider extends ProjectResourceScopeProvider {

	interface IProjectSAFEReleaseProvider {
		SAFEReleaseDescriptor get(IProject p);
	}

	static class ProjectSAFEReleaseProvider implements IProjectSAFEReleaseProvider {
		public SAFEReleaseDescriptor get(IProject p) {
			return ISAFEWorkspacePreferences.SAFE_RELEASE.get(p);
		}
	}

	static class ProjectSAFEReleaseCache implements IProjectSAFEReleaseProvider {
		class InvalidationListener implements IResourceChangeListener {
			public void resourceChanged(IResourceChangeEvent event) {
				try {
					IResourceDelta delta = event.getDelta();
					if (delta != null) {
						IResourceDeltaVisitor visitor = new ResourceDeltaVisitor(event.getType(), new DefaultResourceChangeHandler() {
							@Override
							public void handleProjectCreated(int eventType, IProject project) {
								invalidate();
							}

							@Override
							public void handleProjectOpened(int eventType, IProject project) {
								invalidate();
							}

							@Override
							public void handleProjectRenamed(int eventType, IProject oldProject, IProject newProject) {
								invalidate();
							}

							@Override
							public void handleProjectDescriptionChanged(int eventType, IProject project) {
								invalidate();
							}

							@Override
							public void handleProjectSettingsChanged(int eventType, IProject project, Collection<String> preferenceFileNames) {
								invalidate();
							}

							@Override
							public void handleProjectClosed(int eventType, IProject project) {
								invalidate();
							}

							@Override
							public void handleProjectRemoved(int eventType, IProject project) {
								invalidate();
							}
						});

						delta.accept(visitor);
					}
				} catch (CoreException ex) {
					PlatformLogUtil.logAsError(Activator.getDefault(), ex);
				}
			}
		}

		IProjectSAFEReleaseProvider provider = new ProjectSAFEReleaseProvider();

		Map<IProject, SAFEReleaseDescriptor> cache = new WeakHashMap<IProject, SAFEReleaseDescriptor>();

		ProjectSAFEReleaseCache() {
			ResourcesPlugin.getWorkspace().addResourceChangeListener(new InvalidationListener());
		}

		public synchronized SAFEReleaseDescriptor get(IProject p) {
			SAFEReleaseDescriptor projectRelease = cache.get(p);

			if (projectRelease == null) {
				if (cache.containsKey(p)) {
					return null;
				}

				projectRelease = provider.get(p);
				cache.put(p, projectRelease);
			}

			return projectRelease;
		}

		synchronized void invalidate() {
			cache.clear();
		}
	}

	protected ProjectSAFEReleaseCache projectSAFEReleaseCache = new ProjectSAFEReleaseCache();

	public SAFEProjectResourceScopeProvider() {
		ResourceScopeMarkerSynchronizer.INSTANCE.addDelegate(new SAFEResourceScopeMarkerSynchronizerDelegate());
	}

	/*
	 * @see
	 * org.eclipse.sphinx.emf.scoping.ProjectResourceScopeProvider#createScope(org.eclipse.core.resources.IResource)
	 */
	@Override
	protected ProjectResourceScope createScope(IResource resource) {
		SAFEProjectResourceScope scope = new SAFEProjectResourceScope(resource);
		configureScope(scope);
		return scope;
	}

	protected void configureScope(SAFEProjectResourceScope scope) {
		super.configureScope(scope);
		scope.setProjectAutosarReleaseProvider(projectSAFEReleaseCache);
	}

	/*
	 * @see org.eclipse.sphinx.emf.scoping.IResourceScopeProvider#validate(org.eclipse.core.resources.IFile)
	 */
	@Override
	public Diagnostic validate(IFile file) {
		IMetaModelDescriptor effectiveMMDescriptor = MetaModelDescriptorRegistry.INSTANCE.getEffectiveDescriptor(file);

		// only validate model files
		if (effectiveMMDescriptor == null) {
			return Diagnostic.OK_INSTANCE;
		}

		IResourceScope scope = getScope(file);

		// A scope will only be returned if the file belongs to the scope which is not the case if the file's SAFE
		// release does not match the project's SAFE release. For such cases validation yields an error.
		if (scope == null) {
			SAFEReleaseDescriptor projectRelease = ISAFEWorkspacePreferences.SAFE_RELEASE.get(file.getProject());

			if (projectRelease == null) {
				// most likely the project doesn't have the SAFE nature
				// ignore such cases
				return Diagnostic.OK_INSTANCE;
			}

			String msg = NLS.bind(Messages.warning_safeReleaseOfResourceNotCompatibleWithSAFEReleaseOfEnclosingProject,
					new String[] { file.getName(), projectRelease.getName(), effectiveMMDescriptor.getName() });
			return new BasicDiagnostic(Diagnostic.ERROR, Activator.getPlugin().getSymbolicName(), 0, msg, new Object[] { file });
		}

		return Diagnostic.OK_INSTANCE;
	}
}
