/**
 * <copyright>
 * 
 * Copyright (c) 2013 itemis and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *     itemis - Initial API and implementation
 * 
 * </copyright>
 */
package org.safe.ide.util;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.safe.ide.metamodel.SAFEReleaseDescriptor;
import org.safe.ide.internal.Activator;
import org.safe.ide.internal.messages.Messages;
import org.safe.ide.natures.SAFENature;
import org.safe.ide.preferences.ISAFEWorkspacePreferences;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.sphinx.emf.internal.metamodel.InternalMetaModelDescriptorRegistry;
import org.eclipse.sphinx.emf.model.IModelDescriptor;
import org.eclipse.sphinx.emf.model.ModelDescriptorRegistry;
import org.eclipse.sphinx.platform.util.ExtendedPlatform;
import org.eclipse.sphinx.platform.util.PlatformLogUtil;
import org.eclipse.sphinx.platform.util.StatusUtil;

@SuppressWarnings("restriction")
public class SAFEPlatformUtil {
	 /**
     * Collects SAFE projects in one group. This method recursively searches
     * all referenced projects for the current visited project and also the
     * projects that reference the current visited project if flag
     * <code>includeReferencingProjects</code> is set to <code>true</code>.
     * 
     * @param project
     *            The project being visited.
     * @param includeReferencingProjects
     *            Flag indicating if referencing projects must be considered or
     *            not.
     * @param projectGroup
     *            The project group to fill.
     * @param visitedProjects
     *            The already visited projects.
     */
    private static void collectSAFEProjectsInGroup(IProject project,
	    boolean includeReferencingProjects, Set<IProject> projectGroup,
	    Set<IProject> visitedProjects) {
	if (project.isAccessible() && visitedProjects.add(project)) {
	    if (hasSAFENature(project)) {
		projectGroup.add(project);
	    }
	    for (IProject p : ExtendedPlatform
		    .getReferencedProjectsSafely(project)) {
		if (p.isAccessible() && hasSAFENature(p)) {
		    collectSAFEProjectsInGroup(p,
			    includeReferencingProjects, projectGroup,
			    visitedProjects);
		}
	    }
	    if (includeReferencingProjects) {
		for (IProject p : ExtendedPlatform
			.getReferencingProjectsSafely(project)) {
		    if (p.isAccessible() && hasSAFENature(p)) {
			collectSAFEProjectsInGroup(p,
				includeReferencingProjects, projectGroup,
				visitedProjects);
		    }
		}
	    }
	}
    }

    /**
     * Checks if the specified {@linkplain IProject project} has the
     * {@linkplain SAFENature#SAFE_NATURE_ID SAFE nature}; that given
     * project is not supposed to be <code>null</code> and is supposed to be
     * {@linkplain org.eclipse.core.resources.IResource#isAccessible()
     * accessible}. Logs an error is case of {@linkplain CoreException
     * exception}.
     * 
     * @param project
     *            A project that may have the SAFE nature; must not be
     *            <code>null</code> and must be <em>accessible</em>.
     * @return <ul>
     *         <li><tt><b>true</b>&nbsp;&nbsp;</tt> if the specified project has
     *         the SAFE nature;</li>
     *         <li><tt><b>false</b>&nbsp;</tt> otherwise.</li>
     *         </ul>
     */
    public static boolean hasSAFENature(IProject project) {
	Assert.isNotNull(project);
	Assert.isTrue(project.isAccessible());
	boolean hasSAFENature = false;
	try {
	    hasSAFENature = project.hasNature(SAFENature.ID);
	} catch (CoreException ex) {
	    PlatformLogUtil.logAsError(Activator.getDefault(), ex);
	}
	return hasSAFENature;
    }

    /**
     * Computes a group of SAFE projects (see {@link SAFENature})
     * referencing each other. This method will search recursively all
     * referenced projects of the given project and if wanted also the projects
     * that reference the given project.
     * 
     * @param project
     *            A SAFE project whose scope will be computed; must not be
     *            <code>null</code>.
     * @param includeReferencingProjects
     *            If <code>true</code> also includes projects that reference the
     *            given project.
     * @return A collection of SAFE projects which reference one another.
     */
    public static Collection<IProject> getProjectGroup(IProject project,
	    boolean includeReferencingProjects) {
	Assert.isNotNull(project);
	// A set to which all SAFE projects in scope will be added
	Set<IProject> projectGroup = new HashSet<IProject>();
	// Collect SAFE projects in group safely
	collectSAFEProjectsInGroup(project, includeReferencingProjects,
		projectGroup, new HashSet<IProject>());
	return projectGroup;
    }

    public static Collection<IProject> getProjectGroup(
	    IModelDescriptor modelDescriptor) {
	Assert.isNotNull(modelDescriptor);
	// A set to which all SAFE projects in scope will be added
	Set<IProject> projectGroup = new HashSet<IProject>();
	// Collect SAFE projects behind given model descriptor
	IResource root = modelDescriptor.getRoot();
	if (root != null) {
	    IProject project = root.getProject();
	    if (project.isAccessible() && hasSAFENature(project)) {
		projectGroup.add(project);
	    }
	}
	for (IResource resource : modelDescriptor.getReferencedRoots()) {
	    IProject project = resource.getProject();
	    if (project.isAccessible() && hasSAFENature(project)) {
		projectGroup.add(project);
	    }
	}
	return projectGroup;
    }

    /**
     * Converts the given project to an SAFE project and configures it.
     * 
     * @param project
     *            the project to convert.
     * @param safeRelease
     *            the release of the SAFE metamodel that will be used by
     *            this project.
     * @param monitor
     *            a progress monitor.
     */
    @SuppressWarnings("restriction")
	public static IStatus convertToSAFEProject(IProject project,
	    SAFEReleaseDescriptor safeRelease, IProgressMonitor monitor) {
	Assert.isNotNull(project);
	Assert.isNotNull(safeRelease);

	SubMonitor progress = SubMonitor.convert(monitor, 100);
	if (progress.isCanceled()) {
	    throw new OperationCanceledException();
	}

	// Sets the SAFE nature to project
	if (!hasSAFENature(project)) {
	    progress.subTask(Messages.job_AddProjectNatures_SAFE); //$NON-NLS-1$
	    try {
		SAFENature.addTo(project, progress);
	    } catch (CoreException ex) {
		return StatusUtil.createErrorStatus(Activator.getPlugin(), ex);
	    }
	}

	// Adds release descriptor and sets the RESOURCE_VERSION
	ISAFEWorkspacePreferences.RESOURCE_VERSION.set(project,
		safeRelease);
	SAFEReleaseDescriptor currentReleaseDescriptor = ISAFEWorkspacePreferences.SAFE_RELEASE
		.get(project);
	if (currentReleaseDescriptor == null) {
	    ISAFEWorkspacePreferences.SAFE_RELEASE.setInProject(project,
		    safeRelease);
	}

	// Updates cached descriptors by removed the cache in order to avoid
	// obtaining null as model descriptor
	for (IFile file : ExtendedPlatform.getAllFiles(project, false)) {
	    if (SAFEReleaseDescriptor.SAFE_DEFAULT_FILE_EXTENSION
		    .equals(file.getFileExtension())) {
		InternalMetaModelDescriptorRegistry.INSTANCE
			.removeCachedDescriptor(file);
		InternalMetaModelDescriptorRegistry.INSTANCE
			.addCachedDescriptor(file, safeRelease);
	    }
	}

	// Adds model descriptors in registry
	ModelDescriptorRegistry.INSTANCE.addModels(project);
	return Status.OK_STATUS;
    }
}
